@@ -22,14 +22,17 @@ use itertools::Itertools;
 use std::cell::{Cell, RefCell};
 use std::collections::{BTreeMap, HashMap, HashSet, VecDeque};
 use std::convert::TryInto;
+use std::ffi::OsStr;
 use std::fmt::Write as FmtWrite;
 use std::io::{Read, Seek, Write};
 use std::os::unix::ffi::OsStrExt;
 use std::os::unix::fs::MetadataExt;
 use std::os::unix::fs::PermissionsExt;
 use std::path::{Path, PathBuf};
+use std::rc::Rc;
 
 // These chunk parameters could be investigated and tuned.
+
 pub const CHUNK_MIN_SIZE: usize = 256 * 1024;
 pub const CHUNK_MAX_SIZE: usize = 8 * 1024 * 1024;
 
@@ -57,7 +60,7 @@ pub fn open_repository(
             let clock_skew = (resp.unix_now_millis as i64) - chrono::Utc::now().timestamp_millis();
             const MAX_SKEW_MINS: i64 = 15;
             const MAX_SKEW_MILLIS: i64 = MAX_SKEW_MINS * 60 * 1000;
-            if clock_skew > MAX_SKEW_MILLIS || clock_skew < -MAX_SKEW_MILLIS {
+            if !(-MAX_SKEW_MILLIS..=MAX_SKEW_MILLIS).contains(&clock_skew) {
                 // This helps protect against inaccurate item timestamps, which protects users from unintentionally
                 // deleting important backups when deleting based on timestamp queries. Instead they will be notified
                 // of the clock mismatch as soon as we know about it.
@@ -82,7 +85,7 @@ pub fn init_repository(
     }
 }
 
-type FileActionLogFn = std::rc::Rc<dyn Fn(&str) -> Result<(), anyhow::Error>>;
+pub type FileActionLogFn = dyn Fn(&str) -> Result<(), anyhow::Error>;
 
 #[derive(Clone)]
 pub struct SendContext {
@@ -96,17 +99,18 @@ pub struct SendContext {
     pub idx_ectx: crypto::EncryptionContext,
     pub metadata_ectx: crypto::EncryptionContext,
     pub gear_tab: rollsum::GearTab,
-    pub checkpoint_bytes: u64,
+    pub checkpoint_seconds: u64,
     pub want_xattrs: bool,
     pub use_stat_cache: bool,
     pub one_file_system: bool,
-    pub file_action_log_fn: Option<FileActionLogFn>,
+    pub ignore_permission_errors: bool,
+    pub file_action_log_fn: Option<Rc<FileActionLogFn>>,
 }
 
 struct SendSession<'a, 'b, 'c> {
     ctx: SendContext,
     start_time: chrono::DateTime<chrono::Utc>,
-    dirty_bytes: u64,
+    next_sync: std::time::Instant,
     transferred_chunks: u64,
     transferred_bytes: u64,
     added_chunks: u64,
@@ -159,10 +163,9 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
         data: std::vec::Vec<u8>,
     ) -> Result<(), anyhow::Error> {
         write_chunk(self.w, addr, &data)?;
-        self.dirty_bytes += data.len() as u64;
         self.transferred_bytes += data.len() as u64;
         self.transferred_chunks += 1;
-        if self.dirty_bytes > self.ctx.checkpoint_bytes {
+        if (self.transferred_chunks % 16 == 0) && std::time::Instant::now() > self.next_sync {
             self.sync()?;
         }
         Ok(())
@@ -324,6 +327,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                 want_xattrs: self.ctx.want_xattrs,
                 want_sparseness: true,
                 want_hash: false,
+                ignore_permission_errors: self.ctx.ignore_permission_errors,
                 exclusions,
                 exclusion_markers,
             },
@@ -385,7 +389,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                         )?;
                         index_ent.data_hash = cache_entry.hashes[i];
                         index_ent.data_cursor = cache_entry.data_cursors[i];
-                        self.write_idx_ent(&index::VersionedIndexEntry::V4(index_ent))?;
+                        self.write_idx_ent(&index::VersionedIndexEntry::V5(index_ent))?;
                     }
                 }
                 None => {
@@ -443,6 +447,14 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                                     smear_detected = true;
                                     continue 'add_dir_ents;
                                 }
+
+                                (_, Err(err))
+                                    if self.ctx.ignore_permission_errors
+                                        && err.kind() == std::io::ErrorKind::PermissionDenied =>
+                                {
+                                    continue 'add_dir_ents;
+                                }
+
                                 (ent_path, Err(err)) => {
                                     anyhow::bail!("unable to read {}: {}", ent_path.display(), err)
                                 }
@@ -489,7 +501,7 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
                             content_hashes.push(index_ent.data_hash)
                         }
 
-                        self.write_idx_ent(&index::VersionedIndexEntry::V4(index_ent))?;
+                        self.write_idx_ent(&index::VersionedIndexEntry::V5(index_ent))?;
                     }
 
                     if self.send_log_session.is_some() && use_stat_cache && !smear_detected {
@@ -527,7 +539,8 @@ impl<'a, 'b, 'c> SendSession<'a, 'b, 'c> {
             Some(ref mut send_log_session) => send_log_session.borrow_mut().checkpoint()?,
             None => (),
         }
-        self.dirty_bytes = 0;
+        self.next_sync =
+            std::time::Instant::now() + std::time::Duration::from_secs(self.ctx.checkpoint_seconds);
         Ok(())
     }
 
@@ -668,7 +681,8 @@ pub fn send(
     let mut session = SendSession {
         ctx: ctx.clone(),
         start_time,
-        dirty_bytes: 0,
+        next_sync: std::time::Instant::now()
+            + std::time::Duration::from_secs(ctx.checkpoint_seconds),
         transferred_bytes: 0,
         transferred_chunks: 0,
         added_chunks: 0,
@@ -1157,7 +1171,7 @@ fn receive_partial_htree(
     let mut range_groups = data_map
         .data_chunk_ranges
         // Test harsher range splits in debug mode.
-        .chunks(if cfg!(debug_assertions) { 1 } else { 100000 });
+        .chunks(if cfg!(debug_assertions) { 2 } else { 100000 });
 
     let mut ranges = range_groups.next().unwrap();
     write_request_data_ranges(w, ranges)?;
@@ -1225,10 +1239,12 @@ fn receive_partial_htree(
                 }
             };
 
-            // Fast forward until we are at the correct data chunk boundary.
             loop {
-                let num_skipped = tr.fast_forward(range.start_idx.0 - current_data_chunk_idx)?;
-                current_data_chunk_idx += num_skipped;
+                if current_data_chunk_idx < range.start_idx.0 {
+                    let num_skipped =
+                        tr.fast_forward(range.start_idx.0 - current_data_chunk_idx)?;
+                    current_data_chunk_idx += num_skipped;
+                }
                 if let Some(height) = tr.current_height() {
                     if height == 0 && current_data_chunk_idx >= range.start_idx.0 {
                         break;
@@ -1242,10 +1258,6 @@ fn receive_partial_htree(
                 }
             }
 
-            if current_data_chunk_idx != range.start_idx.0 {
-                anyhow::bail!("requested data ranges do not match hash tree accounting, seek overshoot detected")
-            }
-
             while current_data_chunk_idx + (data_addresses.len() as u64) <= range.end_idx.0 {
                 match tr.current_height() {
                     Some(0) => match tr.next_addr() {
@@ -1452,6 +1464,7 @@ pub fn restore_to_local_dir(
                 want_xattrs: false,
                 want_hash: true,
                 one_file_system: false,
+                ignore_permission_errors: false,
             },
         )? {
             let indexed_dir = indexed_dir?;
@@ -1715,11 +1728,11 @@ pub fn restore_to_local_dir(
             }
             if let Some(ref xattrs) = ent.xattrs {
                 for (attr, value) in xattrs.iter() {
-                    match xattr::set(to_ch, attr, value) {
+                    match xattr::set(to_ch, OsStr::from_bytes(attr), value) {
                         Ok(()) => (),
                         Err(err) => anyhow::bail!(
                             "failed to list remove xattr {} from {}: {}",
-                            attr.to_string_lossy(),
+                            String::from_utf8_lossy(attr),
                             to_ch.display(),
                             err
                         ),
@@ -1965,7 +1978,7 @@ pub fn repo_sync(
     progress.set_length(all_missing_chunks.len().try_into().unwrap());
     progress.set_style(
         indicatif::ProgressStyle::default_bar()
-            .template("[{elapsed_precise}] {msg} [{wide_bar}] {pos:>7}/{len:7}")
+            .template("[{elapsed_precise}] {msg} [{wide_bar}] {pos:>10}/{len}")
             .progress_chars("=> "),
     );
 
