@@ -1,18 +1,37 @@
-use serde::{Deserialize, Serialize};
+use super::fsutil;
+use serde::{Deserialize, Deserializer, Serialize, Serializer};
 use std::collections::{BTreeMap, HashMap};
+use std::ffi::OsString;
 use std::io::Write;
+use std::os::unix::ffi::{OsStrExt, OsStringExt};
+use std::path::{Path, PathBuf};
 
-pub type Xattrs = BTreeMap<String, Vec<u8>>;
+// Deprecated Xattr format for backwards compatibility.
+pub type StringXattrs = BTreeMap<String, Vec<u8>>;
+pub type OsStringXattrs = BTreeMap<OsString, Vec<u8>>;
+pub type Xattrs = BTreeMap<Vec<u8>, Vec<u8>>;
+
+#[derive(Serialize, Deserialize, Debug, Clone, Copy, Eq, PartialEq)]
+pub enum SparseHint {
+    NotSparse,
+    Sparse { min_hole_size: u64 },
+}
 
 #[non_exhaustive]
 #[derive(Serialize, Deserialize, Debug, Clone)]
 pub enum VersionedIndexEntry {
     V1(V1IndexEntry),
     V2(V2IndexEntry),
-    V3(IndexEntry),
+    V3(V3IndexEntry),
+    V4(V4IndexEntry),
+    V5(IndexEntry),
+    // Forward compatibiltiy stubs to improve error messages.
+    Reserved1,
+    Reserved2,
+    Reserved3,
 }
 
-const CURRENT_INDEX_ENTRY_KIND: u8 = 2;
+const CURRENT_INDEX_ENTRY_KIND: u8 = 3;
 
 #[derive(Serialize, Deserialize, Debug, Clone, Copy)]
 pub enum IndexEntryKind {
@@ -55,10 +74,14 @@ pub struct V1IndexEntry {
     pub link_target: Option<String>,
     pub dev_major: serde_bare::Uint,
     pub dev_minor: serde_bare::Uint,
-    pub xattrs: Option<Xattrs>,
+    pub xattrs: Option<StringXattrs>,
     pub data_cursor: AbsoluteDataCursor,
 }
 
+// Deprecated format kept for backwards compatibility.
+// Was deprecated because absolute data cursors broke
+// deduplication, to solve this we had to change them
+// to relative cursors.
 #[derive(Serialize, Deserialize, Debug, Clone)]
 pub struct V2IndexEntry {
     pub path: String,
@@ -76,13 +99,15 @@ pub struct V2IndexEntry {
     pub link_target: Option<String>,
     pub dev_major: serde_bare::Uint,
     pub dev_minor: serde_bare::Uint,
-    pub xattrs: Option<Xattrs>,
+    pub xattrs: Option<StringXattrs>,
     pub data_cursor: AbsoluteDataCursor,
     pub data_hash: ContentCryptoHash,
 }
 
+// Deprecated format kept for backwards compatibility.
+// Was deprecated to allow us to support sparse files.
 #[derive(Serialize, Deserialize, Debug, Clone)]
-pub struct IndexEntry {
+pub struct V3IndexEntry {
     pub path: String,
     pub mode: serde_bare::Uint,
     pub size: serde_bare::Uint,
@@ -98,6 +123,110 @@ pub struct IndexEntry {
     pub link_target: Option<String>,
     pub dev_major: serde_bare::Uint,
     pub dev_minor: serde_bare::Uint,
+    pub xattrs: Option<StringXattrs>,
+    pub data_cursor: RelativeDataCursor,
+    pub data_hash: ContentCryptoHash,
+}
+
+fn serialize_index_path<S>(p: &Path, s: S) -> Result<S::Ok, S::Error>
+where
+    S: Serializer,
+{
+    s.serialize_bytes(p.as_os_str().as_bytes())
+}
+
+fn deserialize_index_path<'de, D>(d: D) -> Result<PathBuf, D::Error>
+where
+    D: Deserializer<'de>,
+{
+    let buf: Vec<u8> = Deserialize::deserialize(d)?;
+    // XXX: error on embedded null?
+    Ok(PathBuf::from(OsString::from_vec(buf)))
+}
+
+fn serialize_opt_index_path<S>(p: &Option<PathBuf>, s: S) -> Result<S::Ok, S::Error>
+where
+    S: Serializer,
+{
+    match p {
+        Some(p) => s.serialize_some(p.as_os_str().as_bytes()),
+        None => s.serialize_none(),
+    }
+}
+
+fn deserialize_opt_index_path<'de, D>(d: D) -> Result<Option<PathBuf>, D::Error>
+where
+    D: Deserializer<'de>,
+{
+    let v: Option<Vec<u8>> = Deserialize::deserialize(d)?;
+    match v {
+        Some(v) => {
+            // XXX: error on embedded null?
+            Ok(Some(PathBuf::from(OsString::from_vec(v))))
+        }
+        None => Ok(None),
+    }
+}
+
+#[derive(Serialize, Deserialize, Debug, Clone)]
+pub struct V4IndexEntry {
+    // The following serialize helpers are needed to work around serde failing on
+    // PathBuf types that are not valid utf8 (which are valid).
+    #[serde(serialize_with = "serialize_index_path")]
+    #[serde(deserialize_with = "deserialize_index_path")]
+    pub path: PathBuf,
+    pub size: serde_bare::Uint,
+    pub mtime: serde_bare::Uint,
+    pub mtime_nsec: serde_bare::Uint,
+    pub ctime: serde_bare::Uint,
+    pub ctime_nsec: serde_bare::Uint,
+    pub ino: serde_bare::Uint,
+    // Group some items together that don't change
+    // much in between files to aid in compression.
+    pub norm_dev: serde_bare::Uint,
+    pub mode: serde_bare::Uint,
+    pub uid: serde_bare::Uint,
+    pub gid: serde_bare::Uint,
+    pub nlink: serde_bare::Uint,
+    // Same as above.
+    #[serde(serialize_with = "serialize_opt_index_path")]
+    #[serde(deserialize_with = "deserialize_opt_index_path")]
+    pub link_target: Option<PathBuf>,
+    pub dev_major: serde_bare::Uint,
+    pub dev_minor: serde_bare::Uint,
+    pub sparse: bool,
+    pub xattrs: Option<OsStringXattrs>,
+    pub data_cursor: RelativeDataCursor,
+    pub data_hash: ContentCryptoHash,
+}
+
+#[derive(Serialize, Deserialize, Debug, Clone)]
+pub struct IndexEntry {
+    // The following serialize helpers are needed to work around serde failing on
+    // PathBuf types that are not valid utf8 (which are valid).
+    #[serde(serialize_with = "serialize_index_path")]
+    #[serde(deserialize_with = "deserialize_index_path")]
+    pub path: PathBuf,
+    pub size: serde_bare::Uint,
+    pub mtime: serde_bare::Uint,
+    pub mtime_nsec: serde_bare::Uint,
+    pub ctime: serde_bare::Uint,
+    pub ctime_nsec: serde_bare::Uint,
+    pub ino: serde_bare::Uint,
+    // Group some items together that don't change
+    // much in between files to aid in compression.
+    pub norm_dev: serde_bare::Uint,
+    pub mode: serde_bare::Uint,
+    pub uid: serde_bare::Uint,
+    pub gid: serde_bare::Uint,
+    pub nlink: serde_bare::Uint,
+    // Same as above.
+    #[serde(serialize_with = "serialize_opt_index_path")]
+    #[serde(deserialize_with = "deserialize_opt_index_path")]
+    pub link_target: Option<PathBuf>,
+    pub dev_major: serde_bare::Uint,
+    pub dev_minor: serde_bare::Uint,
+    pub sparse: bool,
     pub xattrs: Option<Xattrs>,
     pub data_cursor: RelativeDataCursor,
     pub data_hash: ContentCryptoHash,
@@ -154,7 +283,7 @@ impl IndexEntry {
 // Deprecated, kept around for backwards compatibility.
 // It was a mistake to keep absolute data cursors as they
 // encode the entry position which degrades deduplication.
-#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, Copy)]
+#[derive(Serialize, Deserialize, Eq, PartialEq, Debug, Clone, Copy)]
 pub struct AbsoluteDataCursor {
     pub chunk_start_idx: serde_bare::Uint,
     pub chunk_end_idx: serde_bare::Uint,
@@ -162,14 +291,14 @@ pub struct AbsoluteDataCursor {
     pub end_byte_offset: serde_bare::Uint,
 }
 
-#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, Copy)]
+#[derive(Serialize, Deserialize, Eq, PartialEq, Debug, Clone, Copy)]
 pub struct RelativeDataCursor {
     pub chunk_delta: serde_bare::Uint,
     pub start_byte_offset: serde_bare::Uint,
     pub end_byte_offset: serde_bare::Uint,
 }
 
-#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, Copy)]
+#[derive(Serialize, Deserialize, Eq, PartialEq, Debug, Clone, Copy)]
 pub enum ContentCryptoHash {
     None,
     Blake3([u8; 32]),
@@ -301,7 +430,7 @@ impl IndexEntry {
     }
 }
 
-#[derive(Serialize, Deserialize, Debug, Clone, PartialEq)]
+#[derive(Serialize, Deserialize, Debug, Clone, Eq, PartialEq)]
 pub struct HTreeDataRange {
     pub start_idx: serde_bare::Uint,
     pub end_idx: serde_bare::Uint,
@@ -319,70 +448,73 @@ fn add_ent_to_data_map(
     data_chunk_ranges: &mut Vec<HTreeDataRange>,
     incomplete_data_chunks: &mut HashMap<u64, rangemap::RangeSet<usize>>,
 ) {
-    if ent.size.0 == 0 {
-        return;
-    }
-    // Either coalesce the existing range or insert a new range.
-    if !data_chunk_ranges.is_empty()
-        && ((data_chunk_ranges.last().unwrap().end_idx.0 == cur_chunk_idx)
-            || (data_chunk_ranges.last().unwrap().end_idx.0 + 1 == cur_chunk_idx))
-    {
-        data_chunk_ranges.last_mut().unwrap().end_idx.0 =
-            cur_chunk_idx + ent.data_cursor.chunk_delta.0
-    } else {
-        data_chunk_ranges.push(HTreeDataRange {
-            start_idx: serde_bare::Uint(cur_chunk_idx),
-            end_idx: serde_bare::Uint(cur_chunk_idx + ent.data_cursor.chunk_delta.0),
-        })
+    if ent.size.0 > 0 {
+        // Either coalesce the existing range or insert a new range.
+        if !data_chunk_ranges.is_empty()
+            && ((data_chunk_ranges.last().unwrap().end_idx.0 == cur_chunk_idx)
+                || (data_chunk_ranges.last().unwrap().end_idx.0 + 1 == cur_chunk_idx))
+        {
+            data_chunk_ranges.last_mut().unwrap().end_idx.0 =
+                cur_chunk_idx + ent.data_cursor.chunk_delta.0
+        } else {
+            data_chunk_ranges.push(HTreeDataRange {
+                start_idx: serde_bare::Uint(cur_chunk_idx),
+                end_idx: serde_bare::Uint(cur_chunk_idx + ent.data_cursor.chunk_delta.0),
+            })
+        }
     }
 
     if ent.data_cursor.chunk_delta.0 == 0 {
-        let range = ent.data_cursor.start_byte_offset.0 as usize
-            ..ent.data_cursor.end_byte_offset.0 as usize;
-
-        match incomplete_data_chunks.get_mut(&cur_chunk_idx) {
-            Some(range_set) => {
-                range_set.insert(range);
-            }
-            None => {
-                let mut range_set = rangemap::RangeSet::new();
-                range_set.insert(range);
-                incomplete_data_chunks.insert(cur_chunk_idx, range_set);
+        if ent.size.0 > 0 {
+            let range = ent.data_cursor.start_byte_offset.0 as usize
+                ..ent.data_cursor.end_byte_offset.0 as usize;
+            match incomplete_data_chunks.get_mut(&cur_chunk_idx) {
+                Some(range_set) => {
+                    range_set.insert(range);
+                }
+                None => {
+                    let mut range_set = rangemap::RangeSet::new();
+                    range_set.insert(range);
+                    incomplete_data_chunks.insert(cur_chunk_idx, range_set);
+                }
             }
         }
     } else {
         let start_range = ent.data_cursor.start_byte_offset.0 as usize..usize::MAX;
-        let end_range = 0..ent.data_cursor.end_byte_offset.0 as usize;
-
         match incomplete_data_chunks.get_mut(&cur_chunk_idx) {
             Some(range_set) => {
                 range_set.insert(start_range);
             }
             None => {
-                let mut range_set = rangemap::RangeSet::new();
-                range_set.insert(start_range);
-                incomplete_data_chunks.insert(cur_chunk_idx, range_set);
+                if ent.size.0 > 0 {
+                    let mut range_set = rangemap::RangeSet::new();
+                    range_set.insert(start_range);
+                    incomplete_data_chunks.insert(cur_chunk_idx, range_set);
+                }
             }
         }
 
-        if ent.data_cursor.end_byte_offset.0 != 0 {
-            let mut range_set = rangemap::RangeSet::new();
-            range_set.insert(end_range);
-            let old = incomplete_data_chunks
-                .insert(cur_chunk_idx + ent.data_cursor.chunk_delta.0, range_set);
-            // Because our end chunk is a never before seen index, this
-            // range set must be none.
-            assert!(old.is_none());
-        } else {
-            data_chunk_ranges.last_mut().unwrap().end_idx.0 -= 1;
+        if ent.size.0 > 0 {
+            if ent.data_cursor.end_byte_offset.0 != 0 {
+                let end_range = 0..ent.data_cursor.end_byte_offset.0 as usize;
+                let mut range_set = rangemap::RangeSet::new();
+                range_set.insert(end_range);
+                let old = incomplete_data_chunks
+                    .insert(cur_chunk_idx + ent.data_cursor.chunk_delta.0, range_set);
+                // Because our end chunk is a never before seen index, this
+                // range set must be none.
+                assert!(old.is_none());
+            } else {
+                data_chunk_ranges.last_mut().unwrap().end_idx.0 -= 1;
+            }
         }
-    }
 
-    if let Some(range_set) = incomplete_data_chunks.get(&cur_chunk_idx) {
-        if let Some(range) = range_set.get(&(usize::MAX - 1)) {
-            if range.start == 0 {
-                // The range completely covers the chunk, it is no longer incomplete.
-                incomplete_data_chunks.remove(&cur_chunk_idx);
+        if let Some(range_set) = incomplete_data_chunks.get(&cur_chunk_idx) {
+            if let Some(range) = range_set.get(&(usize::MAX - 1)) {
+                if range.start == 0 {
+                    // The range completely covers the chunk, it is no longer incomplete.
+                    incomplete_data_chunks.remove(&cur_chunk_idx);
+                }
             }
         }
     }
@@ -419,7 +551,7 @@ pub fn data_map_for_predicate(
 }
 
 pub fn pick(
-    path: &str,
+    path: &Path,
     file_index: &CompressedIndex,
 ) -> Result<(Option<CompressedIndex>, DataMap), anyhow::Error> {
     let mut cur_chunk_idx = 0;
@@ -442,13 +574,13 @@ pub fn pick(
 
                     let mut sub_index_writer = CompressedIndexWriter::new();
 
-                    ent.path = ".".to_string();
+                    ent.path = ".".into();
                     sub_index_writer.add(&ent);
 
-                    let strip_prefix = if path == "." {
-                        "".to_string()
+                    let prefix_to_strip = if path.as_os_str() == "." {
+                        "".into()
                     } else {
-                        format!("{}/", path)
+                        fsutil::path_raw_join(path, Path::new("/"))
                     };
 
                     cur_chunk_idx += ent.data_cursor.chunk_delta.0;
@@ -459,7 +591,7 @@ pub fn pick(
                         let mut ent = ent?;
 
                         // Match the directory and children.
-                        if !ent.path.starts_with(&strip_prefix) {
+                        if !ent.path.starts_with(&prefix_to_strip) {
                             cur_chunk_idx += ent.data_cursor.chunk_delta.0;
                             if found_children {
                                 // We have processed all children, exit early.
@@ -471,7 +603,11 @@ pub fn pick(
                         }
                         found_children = true;
 
-                        ent.path = ent.path[strip_prefix.len()..].to_string();
+                        ent.path = ent
+                            .path
+                            .strip_prefix(&prefix_to_strip)
+                            .unwrap()
+                            .to_path_buf();
 
                         sub_index_writer.add(&ent);
 
@@ -515,7 +651,7 @@ pub fn pick(
                 }
                 kind => anyhow::bail!(
                     "unable to pick {} - unsupported directory entry type: {:?}",
-                    path,
+                    path.to_string_lossy(),
                     kind
                 ),
             }
@@ -523,11 +659,11 @@ pub fn pick(
 
         cur_chunk_idx += ent.data_cursor.chunk_delta.0;
     }
-    anyhow::bail!("{} not found in content index", path)
+    anyhow::bail!("{} not found in content index", path.to_string_lossy())
 }
 
 pub fn pick_dir_without_data(
-    path: &str,
+    path: &Path,
     file_index: &CompressedIndex,
 ) -> Result<CompressedIndex, anyhow::Error> {
     let mut iter = file_index.iter();
@@ -543,32 +679,36 @@ pub fn pick_dir_without_data(
         match ent.kind() {
             IndexEntryKind::Directory => {
                 let mut sub_index_writer = CompressedIndexWriter::new();
-                ent.path = ".".to_string();
+                ent.path = ".".into();
                 sub_index_writer.add(&ent);
 
-                let strip_prefix = if path == "." {
-                    "".to_string()
+                let prefix_to_strip = if path.as_os_str() == "." {
+                    "".into()
                 } else {
-                    format!("{}/", path)
+                    fsutil::path_raw_join(path, Path::new("/"))
                 };
 
                 for ent in iter {
                     let mut ent = ent?;
-                    if !ent.path.starts_with(&strip_prefix) {
+                    if !ent.path.starts_with(&prefix_to_strip) {
                         continue;
                     }
-                    ent.path = ent.path[strip_prefix.len()..].to_string();
+                    ent.path = ent
+                        .path
+                        .strip_prefix(&prefix_to_strip)
+                        .unwrap()
+                        .to_path_buf();
                     sub_index_writer.add(&ent);
                 }
                 return Ok(sub_index_writer.finish());
             }
             _ => anyhow::bail!(
                 "unable to pick {} in this context, it is not a directory",
-                path,
+                path.to_string_lossy(),
             ),
         }
     }
-    anyhow::bail!("{} not found in content index", path)
+    anyhow::bail!("{} not found in content index", path.to_string_lossy())
 }
 
 // The index can get huge when we have millions of files, so we use a compressed index in memory.
@@ -596,13 +736,27 @@ pub struct CompressedIndexIterator<'a> {
     reader: lz4::Decoder<std::io::Cursor<&'a Vec<u8>>>,
 }
 
+fn migrate_string_xattrs(attrs: StringXattrs) -> Xattrs {
+    attrs
+        .into_iter()
+        .map(|(k, v)| (k.as_bytes().to_vec(), v))
+        .collect()
+}
+
+fn migrate_os_string_xattrs(attrs: OsStringXattrs) -> Xattrs {
+    attrs
+        .into_iter()
+        .map(|(k, v)| (k.as_bytes().to_vec(), v))
+        .collect()
+}
+
 impl<'a> Iterator for CompressedIndexIterator<'a> {
     type Item = Result<IndexEntry, anyhow::Error>;
 
     fn next(&mut self) -> Option<Result<IndexEntry, anyhow::Error>> {
         match serde_bare::from_reader(&mut self.reader) {
             Ok(VersionedIndexEntry::V1(ent)) => Some(Ok(IndexEntry {
-                path: ent.path,
+                path: ent.path.into(),
                 mode: ent.mode,
                 size: ent.size,
                 uid: ent.uid,
@@ -614,10 +768,11 @@ impl<'a> Iterator for CompressedIndexIterator<'a> {
                 norm_dev: ent.dev,
                 ino: ent.ino,
                 nlink: ent.nlink,
-                link_target: ent.link_target,
+                link_target: ent.link_target.map(|x| x.into()),
                 dev_major: ent.dev_major,
                 dev_minor: ent.dev_minor,
-                xattrs: ent.xattrs,
+                sparse: false,
+                xattrs: ent.xattrs.map(migrate_string_xattrs),
                 data_cursor: RelativeDataCursor {
                     chunk_delta: serde_bare::Uint(u64::MAX),
                     start_byte_offset: serde_bare::Uint(u64::MAX),
@@ -626,6 +781,52 @@ impl<'a> Iterator for CompressedIndexIterator<'a> {
                 data_hash: ContentCryptoHash::None,
             })),
             Ok(VersionedIndexEntry::V2(ent)) => Some(Ok(IndexEntry {
+                path: ent.path.into(),
+                mode: ent.mode,
+                size: ent.size,
+                uid: ent.uid,
+                gid: ent.gid,
+                mtime: ent.mtime,
+                mtime_nsec: ent.mtime_nsec,
+                ctime: ent.ctime,
+                ctime_nsec: ent.ctime_nsec,
+                norm_dev: ent.norm_dev,
+                ino: ent.ino,
+                nlink: ent.nlink,
+                link_target: ent.link_target.map(|x| x.into()),
+                dev_major: ent.dev_major,
+                dev_minor: ent.dev_minor,
+                sparse: false,
+                xattrs: ent.xattrs.map(migrate_string_xattrs),
+                data_cursor: RelativeDataCursor {
+                    chunk_delta: serde_bare::Uint(u64::MAX),
+                    start_byte_offset: serde_bare::Uint(u64::MAX),
+                    end_byte_offset: serde_bare::Uint(u64::MAX),
+                },
+                data_hash: ent.data_hash,
+            })),
+            Ok(VersionedIndexEntry::V3(ent)) => Some(Ok(IndexEntry {
+                path: ent.path.into(),
+                mode: ent.mode,
+                size: ent.size,
+                uid: ent.uid,
+                gid: ent.gid,
+                mtime: ent.mtime,
+                mtime_nsec: ent.mtime_nsec,
+                ctime: ent.ctime,
+                ctime_nsec: ent.ctime_nsec,
+                norm_dev: ent.norm_dev,
+                ino: ent.ino,
+                nlink: ent.nlink,
+                link_target: ent.link_target.map(|x| x.into()),
+                dev_major: ent.dev_major,
+                dev_minor: ent.dev_minor,
+                sparse: false,
+                xattrs: ent.xattrs.map(migrate_string_xattrs),
+                data_cursor: ent.data_cursor,
+                data_hash: ent.data_hash,
+            })),
+            Ok(VersionedIndexEntry::V4(ent)) => Some(Ok(IndexEntry {
                 path: ent.path,
                 mode: ent.mode,
                 size: ent.size,
@@ -641,15 +842,15 @@ impl<'a> Iterator for CompressedIndexIterator<'a> {
                 link_target: ent.link_target,
                 dev_major: ent.dev_major,
                 dev_minor: ent.dev_minor,
-                xattrs: ent.xattrs,
-                data_cursor: RelativeDataCursor {
-                    chunk_delta: serde_bare::Uint(u64::MAX),
-                    start_byte_offset: serde_bare::Uint(u64::MAX),
-                    end_byte_offset: serde_bare::Uint(u64::MAX),
-                },
+                sparse: false,
+                xattrs: ent.xattrs.map(migrate_os_string_xattrs),
+                data_cursor: ent.data_cursor,
                 data_hash: ent.data_hash,
             })),
-            Ok(VersionedIndexEntry::V3(ent)) => Some(Ok(ent)),
+            Ok(VersionedIndexEntry::V5(ent)) => Some(Ok(ent)),
+            Ok(_) => Some(Err(anyhow::format_err!(
+                "index entry seems to be from a future version of bupstash"
+            ))),
             Err(serde_bare::error::Error::Io(err))
                 if err.kind() == std::io::ErrorKind::UnexpectedEof =>
             {
@@ -695,7 +896,7 @@ impl Default for CompressedIndexWriter {
     }
 }
 
-pub fn path_cmp(l: &str, r: &str) -> std::cmp::Ordering {
+pub fn path_cmp(l: &Path, r: &Path) -> std::cmp::Ordering {
     // This path comparison is designed such that our
     // 'put' indexer always produces lists of files that are
     // sorted according to it's order. This lets our diff
@@ -719,13 +920,13 @@ pub fn path_cmp(l: &str, r: &str) -> std::cmp::Ordering {
     use std::cmp::Ordering::*;
     if l == r {
         Equal
-    } else if l == "." {
+    } else if l.as_os_str() == "." {
         Less
-    } else if r == "." {
+    } else if r.as_os_str() == "." {
         Greater
     } else {
-        let mut liter = l.split('/');
-        let mut riter = r.split('/');
+        let mut liter = l.iter();
+        let mut riter = r.iter();
         loop {
             match (liter.next(), riter.next()) {
                 (Some(lelem), Some(relem)) => match lelem.cmp(relem) {
@@ -815,20 +1016,22 @@ mod tests {
     fn test_index_path_cmp() {
         use std::cmp::Ordering::*;
 
-        assert!(path_cmp(".", ".") == Equal);
-        assert!(path_cmp(".", "a") == Less);
-        assert!(path_cmp("a", ".") == Greater);
-        assert!(path_cmp("a", "b/c/e") == Less);
-        assert!(path_cmp("b/c/e", "a") == Greater);
-        assert!(path_cmp("a", "b") == Less);
-        assert!(path_cmp("a/b/c", "b") == Greater);
-        assert!(path_cmp("a/b/c", "d/e") == Less);
-        assert!(path_cmp("a/b/a", "a/b/c") == Less);
-        assert!(path_cmp("a/b/c", "a/b/c") == Equal);
-        assert!(path_cmp("b/a/d", "b/z") == Greater);
-        assert!(path_cmp("b/z", "b/a/d",) == Less);
-        assert!(path_cmp("b/z/d", "b/a/d") == Greater);
-        assert!(path_cmp("b/a/d", "b/z/d") == Less);
+        let path_cmp_str = |l: &str, r: &str| path_cmp(Path::new(l), Path::new(r));
+
+        assert!(path_cmp_str(".", ".") == Equal);
+        assert!(path_cmp_str(".", "a") == Less);
+        assert!(path_cmp_str("a", ".") == Greater);
+        assert!(path_cmp_str("a", "b/c/e") == Less);
+        assert!(path_cmp_str("b/c/e", "a") == Greater);
+        assert!(path_cmp_str("a", "b") == Less);
+        assert!(path_cmp_str("a/b/c", "b") == Greater);
+        assert!(path_cmp_str("a/b/c", "d/e") == Less);
+        assert!(path_cmp_str("a/b/a", "a/b/c") == Less);
+        assert!(path_cmp_str("a/b/c", "a/b/c") == Equal);
+        assert!(path_cmp_str("b/a/d", "b/z") == Greater);
+        assert!(path_cmp_str("b/z", "b/a/d",) == Less);
+        assert!(path_cmp_str("b/z/d", "b/a/d") == Greater);
+        assert!(path_cmp_str("b/a/d", "b/z/d") == Less);
 
         let mut v = vec![
             "a/x.txt",
@@ -839,7 +1042,7 @@ mod tests {
             "b/derp.txt",
             "c/d/slurm",
         ];
-        v.sort_by(|l, r| path_cmp(l, r));
+        v.sort_by(|l, r| path_cmp_str(l, r));
         assert_eq!(
             v,
             vec![
